# -*- coding: utf-8 -*-
#
# Copyright © 2014 Geoffrey T. Dairiki <dairiki@dairiki.org>
#
from __future__ import absolute_import

import logging
import math

from pyramid.compat import NativeIO, text_
from pyramid.request import Request
from pyramid import testing
import pytest


@pytest.fixture
def current_request(request):
    r = Request.blank('/')
    testing.setUp(request=r)
    request.addfinalizer(testing.tearDown)
    return r


class TestIntegration(object):
    """ Integration tests.

    """
    @pytest.fixture
    def logstream(self, monkeypatch):
        """ Patch the root logger to log to an in-memory stream.
        """
        import pyramid_log
        logstream = NativeIO()
        root = logging.getLogger()
        monkeypatch.setattr(root, 'handlers', [])
        handler = logging.StreamHandler(logstream)
        fmt = ("%(asctime)s "
               "%(request.method|-)s %(request.path_qs|-)s : "
               "%(message)s")
        formatter = pyramid_log.Formatter(fmt, datefmt="<DATE>")
        handler.setFormatter(formatter)
        root.addHandler(handler)
        return logstream

    def test_with_request(self, logstream, current_request):
        current_request.GET['foo'] = 'bar'
        current_request.path_info = '/path'
        logger = logging.getLogger()
        logger.warn("testing")
        assert logstream.getvalue() == "<DATE> GET /path?foo=bar : testing\n"

    def test_without_request(self, logstream):
        logger = logging.getLogger()
        logger.warn("is this thing on?")
        assert logstream.getvalue() == "<DATE> - - : is this thing on?\n"


@pytest.fixture
def log_record():
    return logging.LogRecord('test', logging.INFO, __file__, 0, '', (), None)


class TestFormatter(object):
    def make_one(self, *args, **kwargs):
        from pyramid_log import Formatter
        return Formatter(*args, **kwargs)

    def test_with_explicit_request(self, log_record):
        log_record.request = Request.blank('/', POST={})
        formatter = self.make_one('%(request.method)s')
        assert formatter.format(log_record) == 'POST'
        assert hasattr(log_record, 'request')

    def test_with_threadlocal_request(self, current_request, log_record):
        formatter = self.make_one('%(request.method)s')
        assert formatter.format(log_record) == 'GET'
        assert not hasattr(log_record, 'request')

    def test_with_no_request(self, log_record):
        formatter = self.make_one('%(request|no request)s')
        assert formatter.format(log_record) == 'no request'
        assert not hasattr(log_record, 'request')

    def test_format_asctime(self, log_record):
        formatter = self.make_one('asctime=%(asctime)s', datefmt="DATEFMT")
        assert formatter.format(log_record) == 'asctime=DATEFMT'

    def test_format_parentheses_in_fallback(self, log_record):
        formatter = self.make_one('%(request.method|(no request))s')
        assert formatter.format(log_record) == '(no request)'

    def test_format_called_with_log_disabled(self, log_record):
        manager = logging.root.manager

        class MockRequest(object):
            @property
            def disable(self):
                return manager.disable
        log_record.request = MockRequest()
        formatter = self.make_one('%(request.disable)s')
        assert formatter.format(log_record) == '%d' % log_record.levelno
        # Check that manager.disable is restored
        assert not manager.disable


class MockObject(object):
    def __init__(self, **kw):
        self.__dict__.update(kw)


class MockWrapper(object):
    def __init__(self, d):
        self.wrapped = d


class TestWrapDict(object):
    def make_one(self, obj, dictwrapper):
        from pyramid_log import _WrapDict
        return _WrapDict(obj, dictwrapper)

    def test_dict(self):
        obj = MockObject()
        proxy = self.make_one(obj, MockWrapper)
        assert type(proxy.__dict__) is MockWrapper
        assert proxy.__dict__.wrapped is obj.__dict__

    def test_getattr(self):
        class Obj(object):
            def m(self):
                return 'foo'
            x = 'bar'
        proxy = self.make_one(Obj(), MockWrapper)
        assert proxy.m() == 'foo'
        assert proxy.x == 'bar'
        with pytest.raises(AttributeError):
            proxy.y

    def test_setattr(self):
        obj = MockObject(x='orig')
        proxy = self.make_one(obj, MockWrapper)
        proxy.x = 'changed'
        assert proxy.x == 'changed'
        assert obj.x == 'changed'
        proxy.y = 'changed'
        assert proxy.y == 'changed'
        assert obj.y == 'changed'

    def test_delattr(self):
        obj = MockObject(x='orig')
        proxy = self.make_one(obj, MockWrapper)
        del proxy.x
        assert obj.__dict__ == {}


EURO_SIGN = text_('\N{EURO SIGN}', 'unicode-escape')


class TestMissing(object):
    def make_one(self, key, fallback=None):
        from pyramid_log import Missing
        return Missing(key, fallback)

    @pytest.mark.parametrize('fallback, expected', [
        (None, "<?attr.name?>"),
        ('foo', "'foo'"),
        (EURO_SIGN, repr(EURO_SIGN)),
        ])
    def test_repr(self, fallback, expected):
        missing = self.make_one("attr.name", fallback)
        assert repr(missing) == expected

    def test_repr_unicode_key(self):
        # carefully constructed to work in python 3.2
        missing = self.make_one(EURO_SIGN)
        assert repr(missing) in (
            r"<?\u20ac?>",              # py2
            "<?%s?>" % EURO_SIGN,       # py3k
            )

    def test_str(self):
        missing = self.make_one('key', "foo")
        assert str(missing) == "foo"

    @pytest.mark.parametrize('fallback, expected', [
        (None, 0),
        ('123', 123),
        ('foo', 0),
        ])
    def test_int(self, fallback, expected):
        missing = self.make_one('somekey', fallback)
        assert int(missing) == expected

    @pytest.mark.parametrize('fallback, expected', [
        ('123', 123.0),
        ('3.14', 3.14),
        ])
    def test_float(self, fallback, expected):
        missing = self.make_one('somekey', fallback)
        assert float(missing) == expected

    def test_float_default_fallback(self):
        missing = self.make_one('somekey', None)
        assert math.isnan(float(missing))

    def test_float_not_a_number(self):
        missing = self.make_one('somekey', 'notanumber')
        assert math.isnan(float(missing))

    @pytest.mark.parametrize('fallback, asint, asfloat', [
        ('foo', '0', 'nan'),
        ('12', '12', '12.00'),
        ])
    def test_format(self, fallback, asint, asfloat):
        missing = self.make_one('key', fallback)
        assert '%s' % missing == fallback
        assert '%r' % missing == repr(fallback)
        assert '%d' % missing == asint
        assert '%.2f' % missing == asfloat

    def test_format_unicode(self):
        missing = self.make_one('key', EURO_SIGN)
        # missing.__str__()
        assert '%s' % missing in (
            r'\u20ac',                  # py2
            EURO_SIGN,                  # py3k
            )
        # This tests missing.__unicode__() under python 2
        assert text_('%s') % missing == EURO_SIGN


class TestDottedLookup(object):
    def make_one(self, dict_):
        from pyramid_log import _DottedLookup
        return _DottedLookup(dict_)

    def test_dotted_attribute_lookup(self):
        d = self.make_one({'a': MockObject(b=MockObject(c='d'))})
        assert d['a.b.c'] == 'd'

    def test_getitem_lookup(self):
        d = self.make_one({'a': {'b': 'c'}})
        assert d['a.b'] == 'c'

    def test_fallback(self):
        d = self.make_one({'a': MockObject(b='x')})
        assert str(d['c|rats']) == 'rats'
        assert str(d['a.c|rats']) == 'rats'
        assert str(d['a.b.c|rats']) == 'rats'

    def test_default_fallback(self):
        d = self.make_one({'a': MockObject(b='x')})
        assert str(d['missing-key']) == '<?missing-key?>'
        assert str(d['a.missing-key']) == '<?a.missing-key?>'



# -*- coding: utf-8 -*-
#
# Copyright © 2014 Geoffrey T. Dairiki <dairiki@dairiki.org>
#
""" A logging formatter which make pyramid request attributes available
for use in log messages.

"""
from __future__ import absolute_import

import logging
import sys
from pyramid.compat import PY3, native_
from pyramid.threadlocal import get_current_request


class Formatter(logging.Formatter):
    ''' A logging formatter which makes attributes of the pyramid
    request available for use in its format string.

    Example usage::

        import logging
        import sys
        import pyramid_log

        handler = logging.StreamHandler(sys.stderr)
        handler.setFormatter(pyramid_log.Formatter(
            "%(asctime)s %(request.unauthenticated_userid|<anonymous>)s "
            "[%(request.client_addr|)s]\\n"
            "  %(request.method|-)s %(request.path_qs|-)\\n"
            "  %(levelname)s %(message)s"))
        root = logging.getLogger()
        root.addHandler(handler)
        root.warning("Say: %s", "howdy")

    which will produce a log message like::

        2014-10-01 17:55:02,001 user.principal [127.0.0.1]
          GET /page?arg=foo
          WARNING Say: howdy

    Passing an explicit request instance
    ====================================

    When generating a log message, the request instance may be passed
    in the ``extra`` dict, e.g.::

        log.warning("Foo: %s", "bar", extra={"request": request})

    If no request is explicitly set,
    :func:`pyramid.threadlocal.get_current_request` is called to
    determine the current request.

    '''
    def format(self, record):
        """ Format the specific record as text.

        """
        has_record = hasattr(record, 'request')
        if not has_record:
            request = get_current_request()
            if request is not None:
                record.request = request

        # magic_record.__dict__ support dotted attribute lookup
        magic_record = _WrapDict(record, _DottedLookup)

        # Disable logging during disable to prevent recursion (in case
        # a logged request property generates a log message)
        save_disable = logging.root.manager.disable
        logging.disable(record.levelno)
        try:
            return logging.Formatter.format(self, magic_record)
        finally:
            logging.disable(save_disable)
            if not has_record and hasattr(record, 'request'):
                del record.request


class _WrapDict(object):
    """ An object proxy which provides a “wrapped” version of the proxied
    object’s ``__dict__``.

    """
    __slots__ = ['_obj', '_dict_wrapper']

    def __init__(self, obj, dict_wrapper):
        self._obj = obj
        self._dict_wrapper = dict_wrapper

    @property
    def __dict__(self):
        return self._dict_wrapper(self._obj.__dict__)

    def __getattr__(self, attr):
        return getattr(self._obj, attr)

    def __setattr__(self, attr, value):
        if attr in self.__slots__:
            object.__setattr__(self, attr, value)
        else:
            setattr(self._obj, attr, value)

    def __delattr__(self, attr):
        return delattr(self._obj, attr)


class Missing(object):
    """ Returned for missing keys.

    This has decent values upon conversion to various types, making
    is usable by printf-style format strings without error.

    """
    def __init__(self, key, fallback=None):
        self.key = key
        self.fallback = fallback

    def __str__(self):
        fallback = self.fallback
        if fallback is None:
            # NB: this differs from __repr__ in that we are allowed to return
            # unicode in python 2
            return '<?%s?>' % self.key
        return fallback

    if not PY3:                         # pragma: no branch
        __unicode__ = __str__

        def __str__(self):
            encoding = sys.getdefaultencoding()
            return self.__unicode__().encode(encoding, 'backslashreplace')

    def __repr__(self):
        fallback = self.fallback
        if fallback is None:
            return '<?%s?>' % native_(self.key, 'ascii', 'backslashreplace')
        return repr(fallback)

    _int_fallback = 0

    def __int__(self):
        fallback = self.fallback
        if fallback is None:
            return self._int_fallback
        try:
            return int(fallback)
        except ValueError:
            return self._int_fallback

    _float_fallback = float('NaN')

    def __float__(self):
        fallback = self.fallback
        if fallback is None:
            return self._float_fallback
        try:
            return float(fallback)
        except ValueError:
            return self._float_fallback


_marker = object()


class _DottedLookup(object):
    """ A dict which supports dotted-key chained lookup, with fallback
    if the lookup fails.

    Example::

        >>> d = _DottedLookup({'a': {'b': 'foo'}})
        >>> d['a.b']                    # same as d['a']['b']
        'foo'

    Fallback values are appended after a pipe (``|``)::

        >>> d['a.b|fb']                 # same as d['a']['b']
        'foo'
        >>> d['a.z|fb']                 # there is no d['a']['z']
        'fb'

    """
    def __init__(self, dict_):
        self.dict = dict_

    def __getitem__(self, key):
        key, pipe, fallback = key.partition('|')
        if not pipe:
            fallback = None
        parts = key.split('.')
        try:
            v = self.dict[parts[0]]
            for i, part in enumerate(parts[1:]):
                a = getattr(v, part, _marker)
                v = a if a is not _marker else v[part]
        except (LookupError, TypeError):
            return Missing(key, fallback)
        else:
            return v



# -*- coding: utf-8 -*-
#
# Copyright © 2014 Geoffrey T. Dairiki <dairiki@dairiki.org>
#
import os
import sys
from setuptools import setup, find_packages

here = os.path.abspath(os.path.dirname(__file__))
README = open(os.path.join(here, 'README.rst')).read()
CHANGES = open(os.path.join(here, 'CHANGES.rst')).read()

version = "0.2.1"

install_requires = [
    'pyramid',
    ]

tests_require = [
    'pytest',
    ]

needs_pytest = {'pytest', 'test', 'ptr'}.intersection(sys.argv)
pytest_runner = ['pytest-runner'] if needs_pytest else []


setup(
    name='pyramid_log',
    version=version,
    description="Include pyramid request attributes in your log messages",
    long_description=README + '\n\n' + CHANGES,
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Intended Audience :: System Administrators",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Framework :: Pyramid",
        "Topic :: Internet :: WWW/HTTP",
        "Topic :: System :: Logging",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: Implementation :: PyPy",
        ],
    keywords="pyramid logging",
    author="Jeff Dairiki",
    author_email="dairiki@dairiki.org",
    url="http://pypi.python.org/pypi/pyramid_log/",
    license="BSD",
    packages=find_packages(),
    # include_package_data=True,
    zip_safe=True,
    setup_requires=pytest_runner,
    install_requires=install_requires,
    tests_require=tests_require,
    )
